//
// de.gyronet.kangaroo.io.AbstractFileDescriptor.java
// kangaroo
//
// Created by Marc Kleine-Budde on 03.28.2003
// Copyright (c) 2003 GYRO net GmbH. All rights reserved.
//

package de.gyronet.kangaroo.io;

import java.io.IOException;
import java.io.InterruptedIOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import javax.microedition.io.Connector;
import javax.microedition.io.StreamConnection;

/**
 * @version $Id: AbstractFileDescriptor.java,v 1.1.1.1 2003/10/31 00:47:08 carter Exp $
 * @author Marc Kleine-Budde
 *
 */
abstract public class AbstractFileDescriptor implements StreamConnection {

    protected int fd = -1;
    protected int mode = 0;

    protected int openCounter = 0;

    protected int inOpenCounter = 0;
    protected int maxInOpen = 1;

    protected int outOpenCounter = 0;
    protected int maxOutOpen = 1;

    protected boolean open = false;

    protected AbstractFileDescriptor() {
    }

    protected void ensureOpen() throws IOException {
        if (!open) {
            throw new IOException("Connection closed");
        }
    }

    private void closeCon() throws IOException {
	    if (--openCounter == 0) {
		close0();
	    }
    }

    public synchronized void close() throws IOException {
	if (open == true) {
	    open = false;

	    closeCon();
	}
    }

    abstract protected void close0() throws IOException;

    private void close(FileDescriptorInputStream in) throws IOException {
	inOpenCounter--;

	closeCon();
    }

    private void close(FileDescriptorOutputStream out) throws IOException {
	outOpenCounter--;

	closeCon();
    }

    public OutputStream openOutputStream() throws IOException {
	ensureOpen();

	if (outOpenCounter == maxOutOpen) {
	    throw new IOException("no more output streams available");
        }

	outOpenCounter++;

        if ((mode & Connector.WRITE) == 0) {
	    throw new IOException("Connection not open for writing");
	}

        return(new FileDescriptorOutputStream(this));
    }

    public DataOutputStream openDataOutputStream() throws IOException {
        return(new DataOutputStream(openOutputStream()));
    }


    public InputStream openInputStream() throws IOException {
	ensureOpen();

	if (inOpenCounter == maxInOpen) {
	    throw new IOException("no more input streams available");
        }

	inOpenCounter++;

        if ((mode & Connector.READ) == 0) {
	    throw new IOException("Connection not open for reading");
	}

        return(new FileDescriptorInputStream(this));
    }

    public DataInputStream openDataInputStream() throws IOException {
        return(new DataInputStream(openInputStream()));
    }


    abstract protected int read() throws IOException;

    protected int read(byte[] buf, int offset, int len) throws IOException {
	return (readBytes(buf, offset, len));
    }

    abstract protected int readBytes(byte[] buf, int off, int len) throws IOException;


    abstract protected void write(int b) throws IOException;

    protected void write(byte[] buf, int off, int len) throws IOException {
	writeBytes(buf, off, len);
    }

    abstract protected void writeBytes(byte[] buf, int off, int len) throws IOException;


    protected synchronized int available() throws IOException {
        return available0();
    }

    abstract protected int available0() throws IOException;


    class FileDescriptorInputStream extends InputStream {

	private AbstractFileDescriptor con;

	FileDescriptorInputStream(AbstractFileDescriptor con) {
	    this.con = con;
	}

	private void ensureOpen() throws InterruptedIOException {
	    if (con == null) {
		throw new InterruptedIOException("Stream closed");
	    }
	}

	public int read() throws IOException {
	    ensureOpen();

	    return (con.read());
	}

	public int read(byte[] buf, int off, int len) throws IOException {
	    ensureOpen();

	    return (con.read(buf, off, len));
	}

	public int available() throws IOException {
	    ensureOpen();

	    return (con.available());
	}

	public void close() throws IOException {
	    ensureOpen();

	    con.close(this);
	    con = null;
	}
    }


    class FileDescriptorOutputStream extends OutputStream {

	private AbstractFileDescriptor con;

	FileDescriptorOutputStream(AbstractFileDescriptor con) {
	    this.con = con;
	}

	private void ensureOpen() throws InterruptedIOException {
	    if (con == null) {
		throw new InterruptedIOException("Stream closed");
	    }
	}

	public void write(int b) throws IOException {
	    ensureOpen();

	    con.write(b);
	}

	public void write(byte[] buf, int off, int len) throws IOException {
	    ensureOpen();

	    con.write(buf, off, len);
	}

	public void close() throws IOException {
	    ensureOpen();

	    con.close();
	    con = null;
	}
    }
}
