/*
 * sch_pktpri.c Sets skb->priority of the packets received.  Simple fifo queue
 *
 *		This program is free software; you can redistribute it and/or
 *		modify it under the terms of the GNU General Public License
 *		as published by the Free Software Foundation; either version
 *		2 of the License, or (at your option) any later version.
 *
 *              Based off of linux/net/sched/sch_fifo.c and changed to do
 *              raw packet modification on 06/12/2003.
 *
 * Authors:     Andrew Stanley-Jones, <asj@instant802.com>
 *              Alexey Kuznetsov, <kuznet@ms2.inr.ac.ru>
 *
 */

#include <linux/config.h>
#include <linux/version.h>
#include <asm/uaccess.h>
#include <asm/system.h>
#include <asm/bitops.h>
#include <linux/types.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/socket.h>
#include <linux/sockios.h>
#include <linux/in.h>
#include <linux/errno.h>
#include <linux/interrupt.h>
#include <linux/if_ether.h>
#include <linux/inet.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/notifier.h>
#include <linux/module.h>
#include <net/ip.h>
#include <net/route.h>
#include <linux/skbuff.h>
#include <net/sock.h>
#include <net/pkt_sched.h>

#if (LINUX_VERSION_CODE > KERNEL_VERSION(2,5,47))
#define NEW_MODULE_CODE
#endif

static int pri_data0 = 16; /* TX DATA 0 queue (lowest priority) */
MODULE_PARM(pri_data0, "i");
MODULE_PARM_DESC(pri_data0,
		 "Priority level for the data0 priority queue desc");

static int pri_data1 = 17; /* TX DATA 1 queue (normal priority) */
MODULE_PARM(pri_data1, "i");
MODULE_PARM_DESC(pri_data1,
		 "Priority level for the data1 priority queue desc");

static int pri_data2 = 18; /* TX DATA 2 queue (high priority) */
MODULE_PARM(pri_data2, "i");
MODULE_PARM_DESC(pri_data2,
		 "Priority level for the data2 priority queue desc");

static int pri_data3 = 19; /* TX DATA 3 queue (highest priority) */
MODULE_PARM(pri_data3, "i");
MODULE_PARM_DESC(pri_data3,
		 "Priority level for the data3 priority queue desc");


struct pktpri_sched_data
{
	unsigned int limit;
	unsigned int priority;
};


#if (LINUX_VERSION_CODE >= KERNEL_VERSION(2,4,22))
static unsigned int
#else
static int
#endif
pktpri_drop(struct Qdisc* sch)
{
	struct sk_buff *skb;

	skb = __skb_dequeue_tail(&sch->q);
	if (skb) {
		sch->stats.backlog -= skb->len;
		kfree_skb(skb);
		return 1;
	}
	return 0;
}

static void
pktpri_reset(struct Qdisc* sch)
{
	skb_queue_purge(&sch->q);
	sch->stats.backlog = 0;
}

static int
pktpri_enqueue(struct sk_buff *skb, struct Qdisc* sch)
{
	struct pktpri_sched_data *q = (struct pktpri_sched_data *)sch->data;

	if (sch->q.qlen <= q->limit) {
                skb->priority = q->priority;
		__skb_queue_tail(&sch->q, skb);
		sch->stats.bytes += skb->len;
		sch->stats.packets++;
		return 0;
	}
	sch->stats.drops++;
#ifdef CONFIG_NET_CLS_POLICE
	if (sch->reshape_fail==NULL || sch->reshape_fail(skb, sch))
#endif
		kfree_skb(skb);
	return NET_XMIT_DROP;
}

static int
pktpri_requeue(struct sk_buff *skb, struct Qdisc* sch)
{
	__skb_queue_head(&sch->q, skb);
	return 0;
}


static struct sk_buff *
pktpri_dequeue(struct Qdisc* sch)
{
	return __skb_dequeue(&sch->q);
}

static int pktpri_init_data0(struct Qdisc *sch, struct rtattr *opt)
{
	struct pktpri_sched_data *q = (void*)sch->data;

        q->priority = pri_data0;
#ifndef NEW_MODULE_CODE
	MOD_INC_USE_COUNT;
#endif
	return 0;
}

static int pktpri_init_data1(struct Qdisc *sch, struct rtattr *opt)
{
	struct pktpri_sched_data *q = (void*)sch->data;

        q->priority = pri_data1;
#ifndef NEW_MODULE_CODE
	MOD_INC_USE_COUNT;
#endif
	return 0;
}

static int pktpri_init_data2(struct Qdisc *sch, struct rtattr *opt)
{
	struct pktpri_sched_data *q = (void*)sch->data;

        q->priority = pri_data2;
#ifndef NEW_MODULE_CODE
	MOD_INC_USE_COUNT;
#endif
	return 0;
}

static int pktpri_init_data3(struct Qdisc *sch, struct rtattr *opt)
{
	struct pktpri_sched_data *q = (void*)sch->data;

        q->priority = pri_data3;
#ifndef NEW_MODULE_CODE
	MOD_INC_USE_COUNT;
#endif
	return 0;
}


static void pktpri_destroy(struct Qdisc* sch)
{
#ifndef NEW_MODULE_CODE
        MOD_DEC_USE_COUNT;
#endif
}


static int pktpri_dump(struct Qdisc *sch, struct sk_buff *skb)
{
	return skb->len;
}


struct Qdisc_ops pktpri_qdisc_ops_data0 =
{
	NULL,
	NULL,
	"pktpri_data0",
	sizeof(struct pktpri_sched_data),

	pktpri_enqueue,
	pktpri_dequeue,
	pktpri_requeue,
	pktpri_drop,

	pktpri_init_data0,
	pktpri_reset,
	pktpri_destroy,
	pktpri_init_data0,

	pktpri_dump,
};


struct Qdisc_ops pktpri_qdisc_ops_data1 =
{
	NULL,
	NULL,
	"pktpri_data1",
	sizeof(struct pktpri_sched_data),

	pktpri_enqueue,
	pktpri_dequeue,
	pktpri_requeue,
	pktpri_drop,

	pktpri_init_data1,
	pktpri_reset,
	pktpri_destroy,
	pktpri_init_data1,

	pktpri_dump,
};


struct Qdisc_ops pktpri_qdisc_ops_data2 =
{
	NULL,
	NULL,
	"pktpri_data2",
	sizeof(struct pktpri_sched_data),

	pktpri_enqueue,
	pktpri_dequeue,
	pktpri_requeue,
	pktpri_drop,

	pktpri_init_data2,
	pktpri_reset,
	pktpri_destroy,
	pktpri_init_data2,

	pktpri_dump,
};


struct Qdisc_ops pktpri_qdisc_ops_data3 =
{
	NULL,
	NULL,
	"pktpri_data3",
	sizeof(struct pktpri_sched_data),

	pktpri_enqueue,
	pktpri_dequeue,
	pktpri_requeue,
	pktpri_drop,

	pktpri_init_data3,
	pktpri_reset,
	pktpri_destroy,
	pktpri_init_data3,

	pktpri_dump,
};


#ifdef MODULE

int init_module(void)
{
        int ret;
	ret = register_qdisc(&pktpri_qdisc_ops_data0);
	if (ret)
                return ret;

	ret = register_qdisc(&pktpri_qdisc_ops_data1);
	if (ret)
                return ret;

	ret = register_qdisc(&pktpri_qdisc_ops_data2);
	if (ret)
                return ret;

	ret = register_qdisc(&pktpri_qdisc_ops_data3);
	return ret;
}

void cleanup_module(void)
{
	unregister_qdisc(&pktpri_qdisc_ops_data0);
	unregister_qdisc(&pktpri_qdisc_ops_data1);
	unregister_qdisc(&pktpri_qdisc_ops_data2);
	unregister_qdisc(&pktpri_qdisc_ops_data3);
}

#endif

MODULE_LICENSE("GPL");
