/*
 * Created on Oct 5, 2003 written by robin (C) 2003 Instant802 Networks Inc
 */
package com.instant802.cldc.LSQL;

import java.io.IOException;

import com.instant802.cldc.DataPath.XMLRepresentationException;
import com.instant802.cldc.net.MalformedHTTPMessageException;
import com.instant802.cldc.net.Socket;
import com.instant802.cldc.net.SocketException;
import com.instant802.cldc.net.SocketTimeoutException;
import com.instant802.cldc.net.TrivialRecordFramer;
import com.instant802.cldc.orchestrator.OrchestratorNode;
import com.instant802.cldc.util.ControlledThread;
import com.instant802.cldc.util.Log;
import com.instant802.cldc.util.ThreadCounter;

/**
 * @author robin
 */
public class LSQLMessageReceiver extends Thread implements ControlledThread {

	private static final String TYPE = "LSQLMessageReceiver";
	private static final int MESSAGE_TIMEOUT = 5000;
	private static final int RECONNECT_DELAY = 5000;
	private static final String HOST = "127.0.0.1";
	private static final int PORT = 8001;
	private final ConfigurationChangeListener listener;
	private Socket socket;
	private TrivialRecordFramer framer;
	private boolean running = true;
	private final OrchestratorNode node;

	public LSQLMessageReceiver(
		OrchestratorNode node,
		ConfigurationChangeListener listener) {
		this.node = node;
		this.listener = listener;
		ThreadCounter.register(this);
	}

	public void finalize() {
		ThreadCounter.deregister(this);
	}

	public void setLocallyModifiedIfClustered()
		throws LSQLException, APConfException {
		if (node.masterCriteria().members() > 1) {
			LSQLQuery query = new LSQLSetLocallyModified();
			LSQLClient client = new LSQLClient(node.schema());
			client.update(query);
		}
	}

	private final void doUpdateQuery( LSQLQuery query ) throws LSQLException, APConfException
	{
		LSQLClient client = new LSQLClient(node.schema());
		client.update(query);
	}
	
	public void clearLocallyModified() throws LSQLException, APConfException {
		doUpdateQuery( new LSQLClearLocallyModified() );
	}

	public void clearConfigRestored() throws LSQLException, APConfException {
		doUpdateQuery( new LSQLClearConfigRestored() );
	}
	
	public void run() {
		while (running) {
			try {
				connect();
				subscribe();
				if (node.adminInterface().addressChanged()) System.exit(0);
				receiveMessages();
			} catch (SocketException e) {
				error(e);
			} catch (IOException e) {
				error(e);
			} catch (XMLRepresentationException e) {
				warning(e);
			} finally {
				disconnect();
			}
			if (running)
				pause(RECONNECT_DELAY);
		}
	}

	private void warning(XMLRepresentationException e) {
		Log.warning(TYPE, e);
	}

	private void pause(int delay) {
		try {
			Thread.sleep(delay);
		} catch (InterruptedException e) {
			error(e);
		}
	}

	private void error(Exception e) {
		Log.error(TYPE, e);
	}

	private synchronized void disconnect() {
		framer = null;
		socket.close();
		socket = null;
		if (Log.DEBUG) debug("disconnected");
	}

	private void receiveMessages()
		throws SocketException, IOException, XMLRepresentationException {
		while (running) {
			try {
				byte[] received = framer.receiveRecord();
				process(received);
			} catch (SocketTimeoutException e) {
				/*
				 * this isn't an error - periodic timeouts are expected. the
				 * running flag can be checked at this point.
				 */
			}
		}
	}

	private void process(byte[] received) throws XMLRepresentationException {
		if (Log.DEBUG) debug("processing: " + new String(received));
		LSQLNotification notification = new LSQLNotification(received);
		if (!notification.origin().equals(LSQLClient.ORIGIN)) {

			try {
				if (notification
					.typeName()
					.equals(LSQLNotification.CONF_CHANGED)) {
					if (Log.DEBUG) Log.notice("configuration has changed at apconfD");
					if (Log.DEBUG) debug(
						"listener is "
							+ listener
							+ " calling handleConfigurationChange");
					listener.handleConfigurationChange();
					if (Log.DEBUG) debug("returned from handleConfigurationChange");
				} else if (
					notification.typeName().equals(LSQLObjectChanged.TYPE)) {
					if (Log.DEBUG) info("received object changed notification from apconfd");
					LSQLObjectChanged objectChanged =
						new LSQLObjectChanged(notification);
					listener.handleObjectChange(
						objectChanged.getObject(node.schema()));
				} else {
					if (Log.DEBUG) info("received notification from apconfD, but it's neither a configuration nor an object changed");
				}
			} catch (XMLRepresentationException e) {
				error(e);
			} catch (IOException e) {
				error(e);
			} catch (MalformedHTTPMessageException e) {
				critical(e);
			} catch (LSQLException e) {
				node.handleLSQLException( e );
			} catch (APConfException e) {
				emergency(e);
			} catch (Exception e) {
				emergency(e);
			}
		}
	}
	
	private void info(String string) {
		Log.info("LSQLMessageReceiver", string );
	}

	private void emergency(Exception e) {
		Log.emergency("LSQLMessageReceiver", e);
	}

	private void critical(Exception e) {
		Log.emergency("LSQLMessageReceiver", e);
	}

	private void subscribe() throws SocketException {
		framer.sendRecord(
//			new LSQLNotifyConfChangeMessage()
//				.asXMLDocument()
//				.toString()
//				.getBytes());
// Alanl non-standard opt				
		new LSQLNotifyConfChangeMessage()
		.asXMLDocument().getBytes());		
		if (Log.DEBUG) debug("subscribed");
	}

	private synchronized void connect() throws SocketException {
		if (Log.DEBUG) debug("connecting");
		socket = new Socket();
		socket.setTimeout(MESSAGE_TIMEOUT);
		socket.connect(HOST, PORT);
		framer = new TrivialRecordFramer(socket);
		if (Log.DEBUG) debug("connected");
	}

	private void debug(String string) {
		Log.debug(TYPE, string);
	}

	public void waitForStop() {
		running = false;
		try {
			join();
		} catch (InterruptedException e) {
			error(e);
		}
	}
}
