/*
 * Created on Nov 1, 2003
 * author: Robin Barooah
 * (C) 2003 Instant802 Networks Inc
 */
package com.instant802.cldc.deviceconfiguration;

import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

import org.kxml2.kdom.Element;

import com.instant802.cldc.DataPath.GenericRecord;
import com.instant802.cldc.DataPath.XMLRecord;
import com.instant802.cldc.DataPath.XMLRepresentationException;
import com.instant802.cldc.util.Log;

/**
 * @author robin
 */
public class ConfigObject extends GenericRecord implements HasPrimaryKey {

	public static String UNDEFINED_FIELD = "";

	private final Schema schema;
	private String type;
	private String name;
	private PrimaryKey primaryKey;


	ConfigObject (Schema schema ) {
		this.schema = schema;
	}

	public ConfigObject(Schema schema, Element element) throws XMLRepresentationException {
		XMLRecord record = new XMLRecord( element );
		makeFromXMLRecord( record );
		type = record.typeName();
		name = record.getNameAttribute();
		this.schema = schema;
	}

	public boolean isShared() {
		return schema.isShared(type);
	}
	/**
	 * this is overridden because generic records don't automatically set a name attribute.
	 */
	protected void startTag( StringBuffer buffer ) {
		if( null != name ) {
			openNameTag( buffer, typeName(), name );
		} else {
			openTag( buffer, typeName() );
		}
	}

	protected void checkType( XMLRecord record ) {
		// DO NOTHING because ConfigObjects are generic.
	}
	
	public String field( String field ) {
		return super.field( field );
	}


	public String typeName() {
		return type;
	}

	/**
	 * Apply a set of field changes to this configuration object.
	 * @param update the set of changes to apply
	 */
	public synchronized void updateFrom(FieldChanges2 update) {
		Hashtable updated = new Hashtable();
		if (Log.DEBUG) debug("deleting and updating fields");
		deleteAndUpdateFields(update, updated);
		if (Log.DEBUG) debug("adding missing fields");
		addMissingFields(update, updated);
		if (Log.DEBUG) debug("setting name attribute");
		setNameAttribute(update.object().getNameAttribute());
	}

	private void addMissingFields(FieldChanges2 update, Hashtable updated) {
		Enumeration e = update.fieldNames();
		while (e.hasMoreElements()) {
			String updateName = (String) e.nextElement();
			if (!updated.containsKey(updateName)) {
				if (Log.DEBUG) debug("setting: "+updateName+" to "+update.field( updateName ));
				field(updateName, update.field(updateName));
			}
		}
	}
	
	private void deleteAndUpdateFields( FieldChanges2 update, Hashtable updated ) {
		Enumeration names = fieldNames();
		while( names.hasMoreElements() ) {
			String name = (String) names.nextElement();
//			if( ! update.hasField( name )) {
//				debug("removing field "+name );
//				removeField( name );
//			} else {
				
				String updateValue = update.field( name );
				if (!updateValue.equals(field( name ))) {
					if (Log.DEBUG) debug("updating "+name+" to "+updateValue );
					field( name, updateValue);
				}
				updated.put( name, name );
//			}
		}
	}

//	private void updateFields( FieldChanges2 update, Hashtable updated ) {
//		Enumeration names = fieldNames();
//		while( names.hasMoreElements() ) {
//			String name = (String) names.nextElement();
//			if( ! update.hasField( name )) {
//				if (Log.DEBUG) debug("removing field "+name );
//				removeField( name );
//			} else {
//				String updateValue = update.field( name );
//				if (!updateValue.equals(field( name ))) {
//					if (Log.DEBUG) debug("updating "+name+" to "+updateValue );
//					field( name, updateValue);
//				}
//				updated.put( name, name );
//			}
//		}
//	}

	public boolean sameType(ConfigObject candidate) {
		return this.typeName().equals(candidate.typeName());
	}

	public String getNameAttribute() {
		return name;
	}

	protected void setNameAttribute(String name) {
		if (null != name) {
			this.name = name;
		}
	}

	/**
	 * return a list of fields which need to be changed in the candidate
	 * to make it match the same set of values as this object.  Fields which are not 
	 * present in this object but are present in the candidate are ignored.
	 * @param candidate the object which would be changed
	 * @param local true if locally written fields as well as shared fields should be adjusted.
	 * @return the list of changes
	 */
	public FieldChanges2 differencesFrom(ConfigObject candidate, boolean local) {
		Enumeration e = fieldNames();
		FieldChanges2 differences = new FieldChanges2(this);
		while (e.hasMoreElements()) {
			String fieldName = (String) e.nextElement();
			if (!candidate.hasField(fieldName)) {
				if( includeField( fieldName, local ) ) {
					if (Log.DEBUG) debug("differences for "+typeName()+" includes "+fieldName+" so adding" );
					differences.addField(fieldName);
				}
			} else if (!field(fieldName).equals(candidate.field(fieldName))) {
				if( includeField( fieldName, local ) ) {
					if (Log.DEBUG) debug("differences for "+typeName()+" includes "+fieldName+" so updating" );
					differences.addField(fieldName);
				}
			}
		}
		return differences;
	}


	/**
	 * return true if the named field should be included in a change set.
	 * @param fieldname the field to evaluate
	 * @param local if true, then changes which are only written locally are also
	 * included.
	 */
	private boolean includeField(String fieldName, boolean local) {
		return schema.isShared( typeName(), fieldName ) || (local && schema.writeLocally( typeName(), fieldName ));
	}

	public boolean hasNameAttribute() {
		return null != getNameAttribute();
	}

	public Enumeration fieldNames() {
		Vector v = new Vector();
		Enumeration e = super.fieldNames();
		while (e.hasMoreElements()) {
			v.addElement(e.nextElement());
		}
		return v.elements();
	}

	public synchronized PrimaryKey primaryKey() {
		if( null == primaryKey ) {
			primaryKey = schema.getPrimaryKey( this );
		}
		return primaryKey;
	}
		
	public synchronized Object keyValue() {
		return primaryKey().value( this );
	}
		
	public boolean sharesField(String fieldName) {
		return schema.isShared(typeName(), fieldName);
	}

	public boolean writeField(String fieldName) {
		return sharesField(fieldName)
			|| schema.writeLocally(typeName(), fieldName);
	}

	public boolean isNull() {
		return false;
	}

	public void setType(String type) {
		this.type = type;
	}
}
