/*
 * Created on Jun 4, 2004
 * author: Robin Barooah
 * Copyright (C) 2004 Instant802 Networks, Inc. All Rights Reserved.
 */
package com.instant802.cldc.net;

import com.instant802.cldc.util.Log;

/**
 * @author robin
 */
public class UnixDatagramSocket extends SocketBase implements DatagramSocket {

	private String defaultDestination = null;
	
	/**
	 * construct a unix domain datagram socket with the local endpoint
	 * having the supplied name.
	 * @param socketName the name used to indentify the local socket
	 */
	public UnixDatagramSocket( String socketName, String defaultDestination ) throws SocketException
		{
		this( socketName );
		this.defaultDestination = defaultDestination;
	}
		
	public UnixDatagramSocket( String localName ) throws SocketException
	{
		constructUnixDomainDatagram();
			try{
				int err = nativeLocalBind(sd, localName );
			if (err != 0) {
				close(); // if we don't close here we potentially leak an SD
				throw new SocketException("couldn't bind datagram socket to local name "+localName+" err="+err );
			}
		} catch ( NullPointerException e ) {
			exitOnNullPointer( e );
		}
	}
	
	/**
	 * Send a byte array to the supplied destination socket.
	 * @param data the data to send
	 * @param destination the name of the destination socket.
	 * @return the number of bytes sent
	 */
	public int sendTo( byte[] data, String destination )
		throws SocketException
		{
			try{
				int sent = 0;
				if( data.length < 1 ) {
					throw new SocketException("cannot send a datagram of size "+data.length );
				}
				try {
					checkOpen();			
				} catch (SocketException e)	{
					throw new SocketException(
							"datagram socket no longer open: "
							+ destination
							+ " errno is: "
							+ -sent);
				}
				
				sent = nativeLocalSendTo(sd, destination, data, data.length);
				
				if (sent < 0)	{
					throw new SocketException(
							"couldn't send datagram to address: "
							+ destination
							+ " errno is: "
							+ -sent);
				}
				return sent;
			} catch ( NullPointerException e ) {
				exitOnNullPointer( e );
				return 0;
			}
			
		}
	/**
	 * receive the next datagram sent to this socket into the supplied buffer
	 * @param data the array to receive data into
	 * @return the number of bytes received.
	 */
	/**
	 * The return 64 bit quantity is as follows: If the bottom 16 bits threated
	 * as a signed 16 bit quantity are > 0 then The top 32 bits are IP address
	 * of sender The next 16 bits are port of sender The bottom 16 bits are
	 * (signed0 length of data Else if bttom 16 bits == 0 then no data was
	 * ready Else an error Alanl
	 */
	public long receiveFrom(byte[] buffer ) throws SocketException
	{
		int buffersize = buffer.length;
		try{
			long res = 0;
			short read = -1;
			int err;
			
			while (read < 0)
			{
				checkOpen();
				res = nativeLocalReceiveFrom(sd, buffer, buffersize);
				
				read = (short) res;
				
				if (read < 0) {
					err = -read;
					throw new SocketException("error during read in receiveFrom:" + err);
				}
			}
			if (0 == read)
				Log.debug("UnixDatagramSocket", "NULL datagram ");
			return res;
		} catch ( NullPointerException e ) {
			exitOnNullPointer( e );
			return 0l;
		}
	}
	
	public Datagram receive(byte[] buf) throws SocketTimeoutException, SocketException
	{
		try{
			long res = receiveFrom(buf );
			Datagram p = new Datagram(buf, (short) res);
			// udp datagrams have a section here extracting the IP address
			// we don't do that for these unix domain datagrams
			return p;
		} catch ( NullPointerException e ) {
			exitOnNullPointer( e );
			return null;
		}
		
	}

	/* (non-Javadoc)
	 * @see com.instant802.cldc.net.DatagramSocket#send(com.instant802.cldc.net.Datagram)
	 */
	public void send(Datagram p) throws SocketException {
		if( defaultDestination != null )
			sendTo( p.getData(), defaultDestination );
		else
			throw new SocketException("send called with default destination not set");
	}
}
