/** 
 * dll.h - A Doubly Linked List implementation.
 * This doubly-linked list implementation achieves the following goals:
 * - No extra structs are needed compared to an integrated implementation.
 *   List management data is stored within each element.
 * - Performs type checking. No generic types or casting to void * is used.
 * - Allows elements to be in multiple lists.
 * Drawbacks:
 * - Implemented using macros, which can make debugging tricky.
 * - Macro expansions result in additional code with each call.
 * These drawbacks could be somewhat overcome with an implementation
 * that used struct offsets to access the list management data.
 */

/**
 * i802_dll_element_members - Declares the members for each element of a DLL.
 * @type: the type of the member.
 * @name: the name of the DLL. Permits an element to be in multiple DLLs.
 * Description: Declares the members dll_@name_next and dll_@name_prev,
 *              of pointer to @type.
 * Example:
 * struct item {
 *      char *item_name;
 *      i802_dll_element_members(struct item, shopping);
 * };
 * Results in:
 * struct item {
 *      char *item_name;
 *      struct item *dll_shopping_next;
 *      struct item *dll_shopping_prev;
 * };
 */
#define i802_dll_element_members(type, name) \
        type *dll_##name##_next; \
	type *dll_##name##_prev

/**
 * i802_dll_list_members - Declares the members for the list struct of a DLL.
 * @type: the type of the member.
 * @name: the name of the DLL.
 * Description: Declares the members dll_@name_first and dll_@name_last,
 *              of pointer to @type. 
 * Example:
 * struct shopping_list {
 *      i802_dll_list_members(struct item, shopping);
 * };
 * Results in:
 * struct shopping_list {
 *      struct item *dll_shopping_first;
 *      struct item *dll_shopping_last;
 * };
 */
#define i802_dll_list_members(type, name) \
	type *dll_##name##_first; \
	type *dll_##name##_last

/**
 * i802_dll_init - Initializes the members of a list struct.
 * @list: a pointer to the list struct
 * @name: the name of the DLL.
 * Description: Sets @list->dll_@name_first and @list->dll_@name_last to NULL.
 * Example:
 * struct shopping_list *l = malloc(sizeof(struct shopping_list));
 * i802_dll_init(l, shopping);
 */
#define i802_dll_init(list, name) \
	(list)->dll_##name##_first = NULL; \
	(list)->dll_##name##_last = NULL

/**
 * i802_dll_empty - Tests if a DLL is empty.
 * @list: a pointer to the list struct.
 * @name: the name of the DLL.
 * Returns: 1 if the list specified by @list and @name is empty; 0 otherwise.
 * Example:
 * if (i802_dll_empty(l, shopping)) { ... }
 */
#define i802_dll_empty(list, name) \
	((list)->dll_##name##_first == NULL)

/**
 * i802_dll_first - Returns the first element of a DLL.
 * @list: a pointer to the list struct.
 * @name: the name of the DLL.
 * Returns: A pointer to the first element of @list/@name, or NULL if empty.
 * Example:
 * char *first_item_name = i802_dll_first(l, shopping)->item_name;
 */
#define i802_dll_first(list, name) \
	((list)->dll_##name##_first)

/**
 * i802_dll_last - Returns the last element of a DLL.
 * @list: a pointer to the list struct.
 * @name: the name of the DLL.
 * Returns: a pointer to the last element of @list/@name, or NULL if empty.
 * Example:
 * char *last_item_name = i802_dll_last(l, shopping)->item_name;
 */
#define i802_dll_last(list, name) \
	((list)->dll_##name##_last)

/**
 * i802_dll_next - Returns the element of a DLL after a given element.
 * @element: a pointer to a given element. Must be in @name.
 * @name:    the name of the DLL.
 * Returns: a pointer to the element following @element in the @name DLL,
 *          or NULL if @element is the last element.
 * Example:
 * struct item *first_item = i802_dll_first(l, shopping);
 * char *second_item_name = i802_dll_next(first_item, shopping)->item_name;
 */
#define i802_dll_next(element, name) \
	((element)->dll_##name##_next)

/**
 * i802_dll_prev - Returns the element of a DLL before a given element.
 * @element: A pointer to a given element. Must be in @name.
 * @name:    the name of the DLL.
 * Returns: A pointer to the element preceeding @element in the @name DLL,
 *          or NULL if @element is the first element.
 * Example:
 * struct item *last_item = i802_dll_last(l, shopping);
 * char *second_last_item_name = i802_dll_prev(last_item, shopping)->item_name;
 */
#define i802_dll_prev(element, name) \
	((element)->dll_##name##_prev)

/**
 * i802_dll_add_last - Adds an element to the end of a DLL.
 * @list:    a pointer to the list struct.
 * @element: a pointer to the element to be added. Must not be in @list/@name.
 * @name:    the name of the DLL.
 * Descrption: Adds @element to the end of @list/@name.
 * Example:
 * struct item *beans = malloc(sizeof(struct item));
 * beans->item_name = "beans";
 * i802_dll_add_last(l, beans, shopping);
 */
#define i802_dll_add_last(list, element, name) \
	(element)->dll_##name##_next = NULL; \
	(element)->dll_##name##_prev = (list)->dll_##name##_last; \
	if ((element)->dll_##name##_prev == NULL) { \
		(list)->dll_##name##_first = element; \
	} else { \
		(element)->dll_##name##_prev->dll_##name##_next = element; \
	} \
	(list)->dll_##name##_last = element

/**
 * i802_dll_add_first - Adds an element to the start of a DLL.
 * @list:    a pointer to the list struct.
 * @element: a pointer to the element to be added. Must not be in @list/@name.
 * @name:    the name of the DLL.
 * Descrption: Adds @element to the start of @list/@name.
 * Example:
 * struct item *candy = malloc(sizeof(struct item));
 * candy->item_name = "candy";
 * i802_dll_add_first(l, candy, shopping);
 */
#define i802_dll_add_first(list, element, name) \
	(element)->dll_##name##_prev = NULL; \
	(element)->dll_##name##_next = (list)->dll_##name##_first; \
	if ((element)->dll_##name##_next == NULL) { \
		(list)->dll_##name##_last = element; \
	} else { \
		(element)->dll_##name##_next->dll_##name##_prev = element; \
	} \
	(list)->dll_##name##_first = element

/**
 * i802_dll_add_after - Adds an item to a DLL after a given element.
 * @list:    a pointer to the list struct.
 * @element: a pointer to the element to be added. Must not be in @list/@name.
 * @name:    the name of the DLL.
 * @after:   a pointer to the given element. Must be in @list/@name.
 * Descrption: Adds @element to @list/@name after @after.
 * Example:
 * struct item *toothbrush = malloc(sizeof(struct item));
 * toothbrush->item_name = "toothbrush";
 * i802_dll_add_after(l, toothbrush, shopping, candy);
 */
#define i802_dll_add_after(list, element, name, after) \
	(element)->dll_##name##_prev = after; \
	(element)->dll_##name##_next = (after)->dll_##name##_next; \
	if ((element)->dll_##name##_next == NULL) { \
		(list)->dll_##name##_last = element; \
	} else { \
		(element)->dll_##name##_next->dll_##name##_prev = element; \
	} \
	(after)->dll_##name##_next = element

/**
 * i802_dll_add_before - Adds an item to a DLL before a given element.
 * @list:    a pointer to the list struct.
 * @element: a pointer to the element to be added. Must not be in @list/@name.
 * @name:    the name of the DLL.
 * @before:  a pointer to the given element. Must be in @list/@name.
 * Descrption: Adds @element to @list/@name before @before.
 * Example:
 * struct item *beano = malloc(sizeof(struct item));
 * beano->item_name = "beano";
 * i802_dll_add_before(l, beano, shopping, beans);
 */
#define i802_dll_add_before(list, element, name, before) \
	(element)->dll_##name##_next = before; \
	(element)->dll_##name##_prev = (before)->dll_##name##_prev; \
	if ((element)->dll_##name##_prev == NULL) { \
		(list)->dll_##name##_first = element; \
	} else { \
		(element)->dll_##name##_prev->dll_##name##_next = element; \
	} \
	(before)->dll_##name##_prev = element

/**
 * i802_dll_remove - Removes an item from a DLL.
 * @list:    a pointer to the list struct.
 * @element: a pointer to the element to be removed. Must be in @list/@name.
 * @name:    the name of the DLL.
 * Descrption: Removes @element from @list/@name.
 * Example:
 * i802_dll_remove(l, beano, shopping); // It's good for the heart.
 */
#define i802_dll_remove(list, element, name) \
	if ((element)->dll_##name##_next == NULL) { \
		(list)->dll_##name##_last = (element)->dll_##name##_prev; \
	} else { \
		(element)->dll_##name##_next->dll_##name##_prev = \
			(element)->dll_##name##_prev; \
	} \
	if ((element)->dll_##name##_prev == NULL) { \
		(list)->dll_##name##_first = (element)->dll_##name##_next; \
	} else { \
		(element)->dll_##name##_prev->dll_##name##_next = \
			(element)->dll_##name##_next; \
	} \
	(element)->dll_##name##_next = NULL; \
	(element)->dll_##name##_prev = NULL

/**
 * i802_dll_move_after - Moves an item in a DLL after given element.
 * @list:    a pointer to the list struct.
 * @element: a pointer to the element to be moved. Must be in @list/@name.
 * @name:    the name of the DLL.
 * @after:   a pointer to the given element. Must be in @list/@name.
 * Descrption: Moves @element in @list/@name after @after.
 * Example:
 * i802_dll_move_after(l, candy, shopping, toothbrush); // Late night treat. 
 */
#define i802_dll_move_after(list, element, name, after) \
	i802_dll_remove(list, element, name); \
	i802_dll_add_after(list, element, name, after)

/**
 * i802_dll_move_before - Moves an item in a DLL before given element.
 * @list:    a pointer to the list struct.
 * @element: a pointer to the element to be moved. Must be in @list/@name.
 * @name:    the name of the DLL.
 * @before:  a pointer to the given element. Must be in @list/@name.
 * Descrption: Moves @element in @list/@name before @before.
 * Example:
 * i802_dll_move_before(l, candy, shopping, toothbrush); // Be responsible.
 */
#define i802_dll_move_before(list, element, name, before) \
	i802_dll_remove(list, element, name); \
	i802_dll_add_before(list, element, name, before)
